#!/usr/bin/ruby

require 'RMagick'
include Math
include Magick

class Array
    def any?(&b)
	result = false
	each{|*i| result ||= b.call(*i)}
	result
    end
end

PI_180 = PI / 180.0

module Projection
    class Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    @xsize = xsize.to_f
	    @ysize = ysize.to_f
	    @north = north.to_f
	    @south = south.to_f
	    @west  = west.to_f
	    @east  = east.to_f
	    @xmul = @xsize / (@east  - @west)
	    @xadd = -@west
	end
	def point_on_map?(x,y)
	    x >= @west && x <= @east && y >= @north && y <= @south
	end
	def bb_on_map?(l,r,u,d)
	    r >= @west && l <= @east && u >= @south && d <= @north
	end
	def cmd
	    "#{cname} #{@xsize} #{@ysize} #{@north} #{@south} #{@west} #{@east}"
	end    
    end
    class Equidistant < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (@south - @north)
	    @yadd = -@north
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (lat  + @yadd) * @ymul
	    ]
	end
	def cname
	    "d"
	end
    end
    class Equalarea < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (sin (@south * PI_180) - sin(@north * PI_180))
	    @yadd = -sin(@north * PI_180)
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (sin (lat * PI_180) + @yadd) * @ymul
	    ]
	end
	def cname
	    "a"
	end
    end
    class Mercator < Cylindrical
	def forw(lat)
	    lat = (89 + lat)/2 if lat > 89
	    lat = (-89 + lat)/2 if lat < -89
	    log(tan(lat * PI_180) + 1/cos(lat * PI_180))
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (forw (lat)    - forw(@north))/
	     (forw (@south) - forw(@north)) * @ysize
	    ]
	end
	def cname
	    "m"
	end
    end
end

class Box < Struct.new(:l,:r,:u,:d)
    def disjoint?(b)
	b.r <= l || b.l >= r || b.u >= d || b.d <= u
    end
    def overlap(b)
	[r - b.l, b.r - l].max * [d - b.u, b.d - u].max
    end
    def +(b)
	Box.new([l,b.l].min, [r,b.r].max, [u,b.u].max, [d,b.d].min)
    end
    def mx
	(l + r) / 2
    end
    def my
	(u + d) / 2
    end
end

class CityDot < Struct.new(:x,:y,:overlap_range)
    def box
	@box ||= Box.new(x-overlap_range, x+overlap_range, y-overlap_range, y+overlap_range)
	@box
    end
end

class CityLabel < Struct.new(:rx,:ry,:x,:y,:w,:h,:label,:offmax)
    def box
	l = x + rx * (w / 2 + offmax) - w/2
        r = l + w
        u = y + ry * (h / 2 + offmax) - h/2
        d = u + h
	Box.new (l,r,u,d)
    end
end

class Labels
    def initialize(image)
	@image = image
	@text = Draw.new
	@text.text_antialias = true
	@text.pointsize = 12
	@text.fill = 'white'
	@labels = []
	@citydots = []
    end
    def improve_placement(temp)
	buckets=[]
	@labels.each_with_index {|l,i|
	    b = l.box
	    conflicts = []
	    buckets.each_with_index{|(bb,labels),j|
		if not b.disjoint?(bb) && labels.any?{|k| not b.disjoint?(@labels[k].box)}
		    conflicts.push j
		end
	    }
	    lss=[i]
	    dsf=0
	    conflicts.each {|j|
	        bb,ls = buckets.delete_at (j-dsf)
		dsf += 1
		b   += bb
	        lss += ls
	    }
	    buckets.push [b, lss]
	}
	vx=[]
	vy=[]
	buckets.each{|b,ls|
	    ls.each {|i| ls.each{|j|
		next if i == j
		l = @labels[i].box
		q = @labels[j].box
		unless l.disjoint? q
		    ivx = (l.mx - q.mx)
		    ivy = (l.my - q.my)
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    o = l.overlap q
		    vx[i] ||= 0.0
		    vy[i] ||= 0.0
		    vx[i] += o * ivx / ivn
		    vy[i] += o * ivy / ivn
		end
	    }}
	}
	@labels.each_with_index{|el,i|
	    b = el.box
	    rx = el.rx
	    ry = el.ry
	    vx[i] ||= 0.0
	    vy[i] ||= 0.0
	    @citydots.each{|c|
		unless b.disjoint? c.box
		    ivx = b.mx - c.x
		    ivy = b.mx - c.y
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    vx[i] += ivx / ivn
		    vy[i] += ivy / ivn
		end
	    }
	    n = sqrt (vx[i] * vx[i] + vy[i] * vy[i])
	    if n == 0
		if rx.abs < ry.abs
		    rx = (rx > 0) ? [rx-temp, 0].max : [rx+temp, 0].min
		else
		    ry = (ry > 0) ? [ry-temp, 0].max : [ry+temp, 0].min
		end
	    else
	        rx = rx + temp * vx[i]/n
		ry = ry + temp * vy[i]/n
		rx =  1 if rx >  1
		rx = -1 if rx < -1
		ry =  1 if ry >  1
		ry = -1 if ry < -1
	    end
	    @labels[i].rx = rx
	    @labels[i].ry = ry
	}
    end
    def draw
	10.times {improve_placement 0.5}
	10.times {improve_placement 0.4}
	10.times {improve_placement 0.3}
	10.times {improve_placement 0.2}
	10.times {improve_placement 0.1}
	10.times {improve_placement 0.05}
	@labels.each {|l|
	    b = l.box
	    @text.annotate(@image, 0, 0, b.l, b.d, l.label)
	}
    end
    def add_city(size,name,long,lat)
	x,y = $proj.project (long,lat)
	if size == "b"
	    size          = 4
	    overlap_range = 6
	    offmax        = 8
	else
	    size          = 2.5
	    overlap_range = 3
	    offmax        = 5
	end

	dot = Draw.new
	dot.stroke 'none'
        dot.fill 'white'
#	dot.antialias true
	dot.circle(x, y, x, y + size)
        dot.draw @image
	
	metrics = @text.get_type_metrics(@image,name)
	@labels.push CityLabel.new(1, 0, x, y, metrics.width, metrics.ascent, name, offmax)
	@citydots.push CityDot.new(x, y, overlap_range)
    end
end

projection_type = STDIN.readline.chomp
map_x           = STDIN.readline.to_i
map_y           = STDIN.readline.to_i
projection_n    = STDIN.readline.to_f
projection_s    = STDIN.readline.to_f
projection_w    = STDIN.readline.to_f
projection_e    = STDIN.readline.to_f
outfile		= STDIN.readline.chomp

case projection_type
when 'equalarea','a'
    $proj = Projection::Equalarea.new  (map_x, map_y, projection_n, projection_s, projection_w, projection_e)
when 'equidistant','d'
    $proj = Projection::Equidistant.new(map_x, map_y, projection_n, projection_s, projection_w, projection_e)
when 'mercator','m'
    $proj = Projection::Mercator.new   (map_x, map_y, projection_n, projection_s, projection_w, projection_e)
else
    STDERR.print "Unknown projection #{projection_type}\n"
    exit 1
end

background      = STDIN.readline.chomp

case background
when 'satellite','bluemarble'
    type_bm = true
when 'political','polygons'
    type_bm = false
else
    STDERR.print "Unknown map type: #{background}\n"
    exit 1
end

rivers = false
STDIN.each{|line|
    line.chomp!
    case line
    when "rivers"
	rivers = true
    else
	STDERR.print "Unknown line: #{line}\n"    
    end
}

if type_bm
    system ("./project " + $proj.cmd + " >tmp-img.ppm")
    image = ImageList.new("tmp-img.ppm")

    open("pp/borders") {|fn| fn.each{|line|
	line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
        next if $1.to_i < 4
        next unless $proj.bb_on_map? ($2.to_f, $3.to_f, $5.to_f, $4.to_f)
        d = Draw.new
        d.stroke_width 1.5
        d.stroke "gray60"
        d.fill "none"
        outline = $6.split(" ")
        outline2 = []
        outline.each {|xy|
	    x,y = xy.split(",")
	    x,y = $proj.project(x.to_f, y.to_f)
	    outline2.push x
	    outline2.push y
	}
        d.polyline(*outline2)
	d.draw image
    }}
else
    lab=[]
    open("pp/lab"){|fn|fn.each{|line|
	line =~ /(\d+) (\S+)/
        lab[$1.to_i]=$2
    }}

    arc=[]
    arcbb=[]
    open("pp/borders"){|fn|fn.each{|line|
	line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
        arc.push $6.split(" ")
	arcbb.push Box.new($2.to_f, $3.to_f, $5.to_f, $4.to_f)
    }}

    image = ImageList.new("in.ppm")

    nr = 0
    open("pp/pal"){|fn|fn.each{|line|
	line.chomp!
        ptx = nil
        pty = nil
        holex = nil
        holey = nil
        outline = []

        nr += 1
        next if nr == 1040

        catch (:dontrender) {
    	    p = line.split (" ").map{|z|z.to_i}
	    box = nil
	    p.each {|num|
		next if num == 0
	        b = arcbb[num.abs - 1]
		if box
		    box += b
		else
		    box = b
	        end
	    }
	    next unless $proj.bb_on_map?(box.l, box.r, box.u, box.d)
	    p.each{|num|
		if num == 0
		    throw :dontrender unless ptx
		    if holex
			outline.push holex
			outline.push holey
		    else
		        holex,holey = ptx,pty
		    end
		else
		    if num > 0
			c = arc[num-1]
		    else
	    		c = arc[-num-1].reverse
		    end
		    c.each {|ptd|
			x,y = ptd.split(",")
			ptx,pty = $proj.project(x.to_f, y.to_f)
		        outline.push ptx
			outline.push pty
		    }
		end
	    }
	    if holex
	        outline.push holex
		outline.push holey
	    end

	    d = Draw.new
	    d.stroke "none"
	    d.fill lab[nr]
	    d.polygon(*outline)
	    d.draw image
	}
    }}
end

open("pp/rivers") {|fn| fn.each{|line|
    line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
    rank = $1.to_i
    break if rank >= 4

    next unless $proj.bb_on_map? ($2.to_f, $3.to_f, $5.to_f, $4.to_f)

    d = Draw.new
    if rank == 1
        d.stroke_width 3
    elsif rank == 2
	d.stroke_width 2
    else rank == 3
        d.stroke_width 1.5
    end
    if type_bm
	d.stroke "#0040FF"
	d.opacity "50%"
    else
	d.stroke "#0030C0"
    end
    d.fill "none"
    outline = $6.split(" ")
    outline2 = []
    outline.map {|xy|
        x,y = xy.split(",")
	x,y = $proj.project(x.to_f, y.to_f)
        outline2.push x
        outline2.push y
    }
    d.polyline(*outline2)
    d.draw image
}}


labels=Labels.new(image)
fn=open("cities2.pl")
fn.each {|line|
    line =~ /^(.)\s+(.*?)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)/
    labels.add_city($1, $2, $4.to_f, $3.to_f)
}
labels.draw

if type_bm
    text = Draw.new
    text.text_antialias = true

    text.fill = "#B350B350FFFF8000"
    text.font_weight = BoldWeight

    x,y = $proj.project(19.5,51.3)
    label = "POLSKA"
    text.pointsize = 72
    metrics = text.get_type_metrics(image, label)    
    text.annotate(image, 0, 0, x - metrics.width/2, y - metrics.ascent/2, label)

    x,y = $proj.project(15.5,49.3)
    label = "CZECHY"
    text.pointsize = 36
    metrics = text.get_type_metrics(image, label)    
    text.annotate(image, 0, 0, x - metrics.width/2, y - metrics.ascent/2, label)
end

image.write(outfile)
